package io.p2vman.vmodextras.cc.peripheral.net;

import dan200.computercraft.api.lua.LuaException;
import dan200.computercraft.api.lua.LuaFunction;
import dan200.computercraft.api.lua.LuaValues;
import dan200.computercraft.api.peripheral.IComputerAccess;
import dan200.computercraft.api.peripheral.IPeripheral;
import io.p2vman.vmodextras.cc.net.*;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;
import net.minecraft.world.level.Level;
import net.minecraft.world.phys.Vec3;

import javax.annotation.Nonnull;
import java.util.Set;

public abstract class NetworkModemPeripheral implements IPeripheral, IPacketSender, IPacketReceiver
{
    private INetwork network;
    private final Set<IComputerAccess> computers = new ObjectOpenHashSet<>(1);
    private final ModemState state;

    protected NetworkModemPeripheral(ModemState state)
    {
        this.state = state;
    }

    public ModemState getModemState()
    {
        return state;
    }

    private synchronized void setNetwork( INetwork network )
    {
        if( this.network == network ) return;

        // Leave old network
        if( this.network != null ) this.network.removeReceiver( this );

        // Set new network
        this.network = network;

        // Join new network
        if( this.network != null ) this.network.addReceiver( this );
    }

    public void destroy()
    {
        setNetwork( null );
    }

    @Override
    public void receive(@Nonnull NetworkPacket packet )
    {
        if(packet.sender() == this || !state.isOpen(packet.channel())) return;

        synchronized(computers)
        {
            for(IComputerAccess computer : computers )
            {
                computer.queueEvent( "modem_message",
                        computer.getAttachmentName(), packet.channel(), packet.replyChannel(), packet.payload() );
            }
        }
    }

    public void removed() {
        this.setNetwork(null);
    }

    protected abstract INetwork getNetwork();

    @Nonnull
    @Override
    public String getType()
    {
        return "modem";
    }

    private static int parseChannel(int channel) throws LuaException
    {
        if (channel < 0) throw LuaValues.badField("channel", "channel<0", "0");
        return channel;
    }

    @LuaFunction
    public final void open(int channel) throws LuaException
    {
        state.open(parseChannel(channel));
    }

    @LuaFunction
    public final boolean isOpen(int channel) throws LuaException
    {
        return state.isOpen(parseChannel(channel));
    }

    @LuaFunction
    public final void close(int channel) throws LuaException
    {
        state.close(parseChannel(channel));
    }

    @LuaFunction
    public final void closeAll()
    {
        state.closeAll();
    }

    @LuaFunction
    public final int findDomain(String domain) {
        for (IDomainNameHandler handler : this.network.getDomainHandlers()) {
            int a = handler.find(domain);
            if (a != -1) return a;
        }
        return -1;
    }

    @LuaFunction
    public final void transmit(int channel, int replyChannel, Object payload ) throws LuaException
    {
        parseChannel(channel);
        parseChannel(replyChannel);

        Level world = getLevel();
        Vec3 position = getPosition();
        INetwork network = this.network;

        if( world == null || position == null || network == null ) return;

        network.transmit(new NetworkPacket( channel, replyChannel, payload, this ));
    }

    @Override
    public synchronized void attach(@Nonnull IComputerAccess computer)
    {
        synchronized( computers )
        {
            computers.add( computer );
        }

        setNetwork( getNetwork() );
    }

    @Override
    public synchronized void detach(@Nonnull IComputerAccess computer)
    {
        boolean empty;
        synchronized(computers)
        {
            computers.remove(computer);
            empty = computers.isEmpty();
        }

        if(empty) setNetwork(null);
    }

    @Nonnull
    @Override
    public String getSenderID()
    {
        synchronized( computers )
        {
            if( computers.size() != 1 )
            {
                return "unknown";
            }
            else
            {
                IComputerAccess computer = computers.iterator().next();
                return computer.getID() + "_" + computer.getAttachmentName();
            }
        }
    }
}