package io.p2vman.vmodextras.util;

import it.unimi.dsi.fastutil.Pair;
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.fastutil.objects.ObjectObjectImmutablePair;
import lombok.Getter;
import net.minecraft.core.registries.Registries;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.world.item.CreativeModeTab;
import net.minecraft.world.item.CreativeModeTabs;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.entity.BlockEntityType;
import net.minecraftforge.eventbus.api.IEventBus;
import net.minecraftforge.registries.DeferredRegister;
import net.minecraftforge.registries.ForgeRegistries;
import net.minecraftforge.registries.RegistryObject;
import org.jetbrains.annotations.NotNull;

import java.util.*;
import java.util.function.Consumer;
import java.util.function.Supplier;

public class ModRegistry {
    @Getter
    private final String modid;

    public final DeferredRegister<Block> BLOCKS;
    public final DeferredRegister<BlockEntityType<?>> BLOCK_ENTITIES;
    public final Map<RegistryObject<CreativeModeTab>, DeferredRegister<Item>> ITEMS;
    public final DeferredRegister<CreativeModeTab> CREATIVE_MODE_TABS;

    public ModRegistry(String modid) {
        this.modid = modid;
        this.BLOCKS = DeferredRegister.create(ForgeRegistries.BLOCKS, modid);
        this.BLOCK_ENTITIES = DeferredRegister.create(ForgeRegistries.BLOCK_ENTITY_TYPES, modid);
        this.ITEMS = new Object2ObjectArrayMap<>();
        this.CREATIVE_MODE_TABS = DeferredRegister.create(Registries.CREATIVE_MODE_TAB, modid);
    }

    public TabBuilder TabBuilder() {
        return new TabBuilder(this);
    }

    public void register(IEventBus bus) {
        this.BLOCKS.register(bus);
        this.BLOCK_ENTITIES.register(bus);
        this.CREATIVE_MODE_TABS.register(bus);
        for (DeferredRegister<Item> registry : ITEMS.values()) {
            registry.register(bus);
        }
    }



    public static class TabBuilder {
        private final ModRegistry registry;
        public final DeferredRegister<Item> ITEMS;

        public TabBuilder(ModRegistry registry) {
            this.registry = registry;
            this.ITEMS = DeferredRegister.create(ForgeRegistries.ITEMS, registry.modid);
        }

        public TabBuilder push(Consumer<DeferredRegister<Item>> registry) {
            registry.accept(ITEMS);
            return this;
        }

        public RegistryObject<CreativeModeTab> build(String name, ItemStack icon) {
            RegistryObject<CreativeModeTab> tab = this.registry.CREATIVE_MODE_TABS.register(name, () -> CreativeModeTab.builder().withTabsBefore(CreativeModeTabs.COMBAT).icon(() -> icon).displayItems((parameters, output) -> {
                for (RegistryObject<Item> object : ITEMS.getEntries()) {
                    output.accept(object.get());
                }
            }).build());

            this.registry.ITEMS.put(tab, this.ITEMS);
            return tab;
        }
    }


    public enum RegistryEntryType {
        BLOCK,
        ITEM,
        BLOCK_ENTITY
    }

    public static class RegistryContainer implements Iterable<Pair<RegistryEntryType, Object>> {
        private final Object[] values;
        private final String id;
        public RegistryContainer(String id) {
            this.values = new Object[RegistryEntryType.values().length];
            this.id = id;
        }

        public Object get(RegistryEntryType key) {
            return values[key.ordinal()];
        }

        public RegistryObject<? extends BlockEntityType<? extends BlockEntity>> getEntity() {
            return getG(RegistryEntryType.BLOCK_ENTITY);
        }

        public RegistryObject<? extends Block> getBlock() {
            return getG(RegistryEntryType.BLOCK);
        }

        public RegistryObject<? extends Item> getItem() {
            return getG(RegistryEntryType.ITEM);
        }

        public <T> T getG(RegistryEntryType key) {
            return (T) values[key.ordinal()];
        }

        public boolean containsKey(RegistryEntryType key) {
            return values[key.ordinal()] != null;
        }

        public int size() {
            int count = 0;
            for (Object v : values) {
                if (v != null) count++;
            }
            return count;
        }

        public Iterable<RegistryEntryType> keys() {
            RegistryEntryType[] keys = RegistryEntryType.values();
            Set<RegistryEntryType> result = EnumSet.noneOf(RegistryEntryType.class);
            for (int i = 0; i < values.length; i++) {
                if (values[i] != null) result.add(keys[i]);
            }
            return result;
        }

        public RegistryContainer register(Consumer<RegistryContainer> consumer) {
            consumer.accept(this);
            return this;
        }

        public void put(RegistryEntryType key, Object value) {
            values[key.ordinal()] = value;
        }


        public RegistryContainer block(Supplier<Block> supplier) {
            put(RegistryEntryType.BLOCK, supplier);
            return this;
        }

        public RegistryContainer entity(Supplier<BlockEntityType<? extends BlockEntity>> supplier) {
            put(RegistryEntryType.BLOCK_ENTITY, supplier);
            return this;
        }

        public RegistryContainer item(Supplier<Item> supplier) {
            put(RegistryEntryType.ITEM, supplier);
            return this;
        }

        @NotNull
        @Override
        public Iterator<Pair<RegistryEntryType, Object>> iterator() {
            List<Pair<RegistryEntryType, Object>> list = new ObjectArrayList<>();
            for (RegistryEntryType key : keys()) {
                list.add(new ObjectObjectImmutablePair<>(key, get(key)));
            }
            return list.iterator();
        }

        public RegistryContainer register(ModRegistry registry) {
            if (containsKey(RegistryEntryType.BLOCK)) put(RegistryEntryType.BLOCK, registry.BLOCKS.register(id, (Supplier<? extends Block>) get(RegistryEntryType.BLOCK)));
            if (containsKey(RegistryEntryType.BLOCK_ENTITY)) put(RegistryEntryType.BLOCK_ENTITY, registry.BLOCK_ENTITIES.register(id, (Supplier<? extends BlockEntityType<? extends BlockEntity>>) get(RegistryEntryType.BLOCK_ENTITY)));

            return this;
        }

        public RegistryContainer register(TabBuilder builder) {
            if (containsKey(RegistryEntryType.ITEM)) put(RegistryEntryType.ITEM, builder.ITEMS.register(id, (Supplier<? extends Item>) get(RegistryEntryType.ITEM)));

            return this;
        }

        public static RegistryContainer create(String id) {
            return new RegistryContainer(id);
        }
    }
}
